@@ -44,8 +44,8 @@ static void
 generate_key (const char *password,
 	      guchar salt[8],
 	      int iterations,
-	      char key[16],
-	      char iv[16])
+	      guchar key[16],
+	      guchar iv[16])
 {
 	sha256Param sha;
 	guchar digest[32];
@@ -53,7 +53,7 @@ generate_key (const char *password,
 	g_assert (iterations >= 1);
 
 	sha256Reset (&sha);
-	sha256Update (&sha, password, strlen (password));
+	sha256Update (&sha, (unsigned char *)password, strlen (password));
 	sha256Update (&sha, salt, 8);
 	sha256Digest (&sha, digest);
 	iterations--;
@@ -149,7 +149,7 @@ verify_decrypted_buffer (GString *buffer)
 	
 	gnome_keyring_md5_init (&md5_context);
 	gnome_keyring_md5_update (&md5_context,
-				  buffer->str + 16, buffer->len - 16);
+				  (guchar *)buffer->str + 16, buffer->len - 16);
 	gnome_keyring_md5_final (digest, &md5_context);
 	
 	return memcmp (buffer->str, digest, 16) == 0;
@@ -158,10 +158,16 @@ verify_decrypted_buffer (GString *buffer)
 static char *
 get_keyring_dir (void)
 {
-	char *dir;
+	char *dir, *gnome2_dir;
 	
 	dir = g_build_filename (g_get_home_dir (), ".gnome2/keyrings", NULL);
 	if (!g_file_test (dir, G_FILE_TEST_IS_DIR)) {
+		gnome2_dir = g_build_filename (g_get_home_dir (), ".gnome2", NULL);
+		if (!g_file_test (gnome2_dir, G_FILE_TEST_IS_DIR)) {
+			mkdir (gnome2_dir, S_IRWXU);
+		}
+		g_free (gnome2_dir);
+		
 		if (mkdir (dir, S_IRWXU) < 0) {
 			g_warning ("unable to create keyring dir");
 		}
@@ -304,7 +310,7 @@ generate_file (GString *buffer, GnomeKeyring *keyring)
 	gnome_keyring_proto_add_uint32 (buffer, flags);
 	gnome_keyring_proto_add_uint32 (buffer, keyring->lock_timeout);
 	gnome_keyring_proto_add_uint32 (buffer, keyring->hash_iterations);
-	g_string_append_len (buffer, keyring->salt, 8);
+	g_string_append_len (buffer, (char *)keyring->salt, 8);
 
 	/* Reserved: */
 	for (i = 0; i < 4; i++) {
@@ -330,7 +336,7 @@ generate_file (GString *buffer, GnomeKeyring *keyring)
 
 	/* Encrypted data: */
 	to_encrypt = g_string_new (NULL);
-	g_string_append_len (to_encrypt, digest, 16); /* Space for hash */
+	g_string_append_len (to_encrypt, (char *)digest, 16); /* Space for hash */
 
 	if (!generate_encrypted_data (to_encrypt, keyring)) {
 		g_string_free (to_encrypt, TRUE);
@@ -344,7 +350,7 @@ generate_file (GString *buffer, GnomeKeyring *keyring)
 
 	gnome_keyring_md5_init (&md5_context);
 	gnome_keyring_md5_update (&md5_context,
-				  to_encrypt->str + 16, to_encrypt->len - 16);
+				  (guchar *)to_encrypt->str + 16, to_encrypt->len - 16);
 	gnome_keyring_md5_final (digest, &md5_context);
 	memcpy (to_encrypt->str, digest, 16);
 	
@@ -359,6 +365,19 @@ generate_file (GString *buffer, GnomeKeyring *keyring)
 	return TRUE;
 }
 
+gboolean 
+remove_keyring_file_from_disk (GnomeKeyring *keyring)
+{
+	int res;
+
+	g_assert (keyring->file != NULL);
+	res = unlink (keyring->file);
+		
+	return (res == 0);
+}
+
+
+
 static int
 write_all (int fd, const char *buf, size_t len)
 {
@@ -760,8 +779,9 @@ update_keyring_from_disk (GnomeKeyring *keyring,
 {
 	struct stat statbuf;
 	GString buffer;
-	char *contents;
+	char *contents = NULL;
 	gsize len;
+	gboolean success = FALSE;
 
 	if (keyring->file == NULL) {
 		return TRUE;
@@ -776,19 +796,17 @@ update_keyring_from_disk (GnomeKeyring *keyring,
 	}
 	keyring->file_mtime = statbuf.st_mtime;
 
-
 	if (!g_file_get_contents (keyring->file,
 				  &contents, &len, NULL)) {
 		return FALSE;
 	}
+
 	buffer.str = contents;
 	buffer.len = buffer.allocated_len = len;
+	success = update_keyring_from_data (keyring, &buffer);
+	g_free (contents);
 
-	if (!update_keyring_from_data (keyring, &buffer)) {
-		return FALSE;
-	}
-	
-	return TRUE;
+	return success;
 }
 
 void
@@ -864,6 +882,7 @@ update_keyrings_from_disk (void)
 	dirname = get_keyring_dir ();
 
 	if (stat (dirname, &statbuf) < 0) {
+		g_free (dirname);
 		return;
 	}
 	if (statbuf.st_mtime == keyring_dir_mtime) {
