@@ -36,6 +36,7 @@
 #include <string.h>
 #include <sys/types.h>
 #include <sys/socket.h>
+#include <sys/uio.h>
 #include <sys/un.h>
 #include <stdarg.h>
 
@@ -253,11 +254,39 @@ write_credentials_byte_sync (int socket)
 {
   char buf;
   int bytes_written;
+#if defined(HAVE_CMSGCRED) && (!defined(LOCAL_CREDS) || defined(__FreeBSD__))
+  union {
+	  struct cmsghdr hdr;
+	  char cred[CMSG_SPACE (sizeof (struct cmsgcred))];
+  } cmsg;
+  struct iovec iov;
+  struct msghdr msg;
+#endif
+
+  buf = 0;
+#if defined(HAVE_CMSGCRED) && (!defined(LOCAL_CREDS) || defined(__FreeBSD__))
+  iov.iov_base = &buf;
+  iov.iov_len = 1;
+
+  memset (&msg, 0, sizeof (msg));
+  msg.msg_iov = &iov;
+  msg.msg_iovlen = 1;
+
+  msg.msg_control = (caddr_t) &cmsg;
+  msg.msg_controllen = CMSG_SPACE (sizeof (struct cmsgcred));
+  memset (&cmsg, 0, sizeof (cmsg));
+  cmsg->cmsg_len = CMSG_LEN (sizeof (struct cmsgcred));
+  cmsg->cmsg_level = SOL_SOCKET;
+  cmsg->cmsg_type = SCM_CREDS;
+#endif
 
  again:
 
-  buf = 0;
+#if defined(HAVE_CMSGCRED) && (!defined(LOCAL_CREDS) || defined(__FreeBSD__))
+  bytes_written = sendmsg (socket, &msg, 0);
+#else
   bytes_written = write (socket, &buf, 1);
+#endif
 
   if (bytes_written < 0 && errno == EINTR)
     goto again;
@@ -275,11 +304,39 @@ write_credentials_byte (GnomeKeyringOperation *op)
 {
   char buf;
   int bytes_written;
+#if defined(HAVE_CMSGCRED) && (!defined(LOCAL_CREDS) || defined(__FreeBSD__))
+  union {
+	  struct cmsghdr hdr;
+	  char cred[CMSG_SPACE (sizeof (struct cmsgcred))];
+  } cmsg;
+  struct iovec iov;
+  struct msghdr msg;
+#endif
+
+  buf = 0;
+#if defined(HAVE_CMSGCRED) && (!defined(LOCAL_CREDS) || defined(__FreeBSD__))
+  iov.iov_base = &buf;
+  iov.iov_len = 1;
+
+  memset (&msg, 0, sizeof (msg));
+  msg.msg_iov = &iov;
+  msg.msg_iovlen = 1;
+
+  msg.msg_control = (caddr_t) &cmsg;
+  msg.msg_controllen = CMSG_SPACE (sizeof (struct cmsgcred));
+  memset (cmsg, 0, sizeof (cmsg));
+  cmsg->cmsg_len = CMSG_LEN (sizeof (struct cmsgcred));
+  cmsg->cmsg_level = SOL_SOCKET;
+  cmsg->cmsg_type = SCM_CREDS;
+#endif
 
  again:
 
-  buf = 0;
+#if defined(HAVE_CMSGCRED) && (!defined(LOCAL_CREDS) || defined(__FreeBSD__))
+  bytes_written = sendmsg (op->socket, &msg, 0);
+#else
   bytes_written = write (op->socket, &buf, 1);
+#endif
 
   if (bytes_written < 0 && errno == EINTR)
     goto again;
@@ -441,7 +498,10 @@ run_sync_operation (GString *buffer,
 	GnomeKeyringResult res;
 	int socket;
 	guint32 packet_size;
-	
+
+	g_assert (buffer != NULL);
+	g_assert (receive_buffer != NULL);
+
 	socket = connect_to_daemon (FALSE);
 	if (socket < 0) {
 		return GNOME_KEYRING_RESULT_NO_KEYRING_DAEMON;
@@ -585,7 +645,7 @@ gnome_keyring_set_default_keyring (const char                             *keyri
 GnomeKeyringResult 
 gnome_keyring_set_default_keyring_sync (const char *keyring)
 {
-	GString *send;
+	GString *send, *receive;
 	GnomeKeyringResult res;
 
 	send = g_string_new (NULL);
@@ -597,8 +657,19 @@ gnome_keyring_set_default_keyring_sync (const char *keyring)
 		return GNOME_KEYRING_RESULT_BAD_ARGUMENTS;
 	}
 
-	res = run_sync_operation (send, NULL);
+	receive = g_string_new (NULL);
+	res = run_sync_operation (send, receive);
 	g_string_free (send, TRUE);
+	if (res != GNOME_KEYRING_RESULT_OK) {
+		g_string_free (receive, TRUE);
+		return res;
+	}
+
+	if (!gnome_keyring_proto_decode_result_reply (receive, &res)) {
+		g_string_free (receive, TRUE);
+		return GNOME_KEYRING_RESULT_IO_ERROR;
+	}
+ 	g_string_free (receive, TRUE);
 
 	return res;
 }
@@ -757,7 +828,7 @@ gnome_keyring_lock_all (GnomeKeyringOperationDoneCallback       callback,
 GnomeKeyringResult 
 gnome_keyring_lock_all_sync (void)
 {
-	GString *send;
+	GString *send, *receive;
 	GnomeKeyringResult res;
 
 	send = g_string_new (NULL);
@@ -768,10 +839,21 @@ gnome_keyring_lock_all_sync (void)
 		return GNOME_KEYRING_RESULT_BAD_ARGUMENTS;
 	}
 
-	res = run_sync_operation (send, NULL);
+ 	receive = g_string_new (NULL);
+	res = run_sync_operation (send, receive);
 	g_string_free (send, TRUE);
-	
-	return res;
+	if (res != GNOME_KEYRING_RESULT_OK) {
+		g_string_free (receive, TRUE);
+		return res;
+	}
+
+        if (!gnome_keyring_proto_decode_result_reply (receive, &res)) {
+                g_string_free (receive, TRUE);
+                return GNOME_KEYRING_RESULT_IO_ERROR;
+        }
+        g_string_free (receive, TRUE);
+
+        return res;
 }
 
 
@@ -806,7 +888,7 @@ GnomeKeyringResult
 gnome_keyring_create_sync (const char *keyring_name,
 			   const char *password)
 {
-	GString *send;
+	GString *send, *receive;
 	GnomeKeyringResult res;
 	
 	send = g_string_new (NULL);
@@ -818,10 +900,21 @@ gnome_keyring_create_sync (const char *keyring_name,
 		return GNOME_KEYRING_RESULT_BAD_ARGUMENTS;
 	}
 	
-	res = run_sync_operation (send, NULL);
+ 	receive = g_string_new (NULL);
+	res = run_sync_operation (send, receive);
 	g_string_free (send, TRUE);
-	
-	return res;
+	if (res != GNOME_KEYRING_RESULT_OK) {
+		g_string_free (receive, TRUE);
+		return res;
+	}
+
+        if (!gnome_keyring_proto_decode_result_reply (receive, &res)) {
+                g_string_free (receive, TRUE);
+                return GNOME_KEYRING_RESULT_IO_ERROR;
+        }
+        g_string_free (receive, TRUE);
+
+        return res;
 }
 
 gpointer
@@ -853,7 +946,7 @@ GnomeKeyringResult
 gnome_keyring_unlock_sync (const char *keyring,
 			   const char *password)
 {
-	GString *send;
+	GString *send, *receive;
 	GnomeKeyringResult res;
 	
 	send = g_string_new (NULL);
@@ -865,9 +958,20 @@ gnome_keyring_unlock_sync (const char *keyring,
 		return GNOME_KEYRING_RESULT_BAD_ARGUMENTS;
 	}
 	
-	res = run_sync_operation (send, NULL);
+ 	receive = g_string_new (NULL);
+	res = run_sync_operation (send, receive);
 	g_string_free (send, TRUE);
-	
+	if (res != GNOME_KEYRING_RESULT_OK) {
+		g_string_free (receive, TRUE);
+		return res;
+	}
+
+        if (!gnome_keyring_proto_decode_result_reply (receive, &res)) {
+                g_string_free (receive, TRUE);
+                return GNOME_KEYRING_RESULT_IO_ERROR;
+        }
+        g_string_free (receive, TRUE);
+
 	return res;
 }
 
@@ -898,7 +1002,7 @@ gnome_keyring_lock (const char                                  *keyring,
 GnomeKeyringResult
 gnome_keyring_lock_sync (const char *keyring)
 {
-	GString *send;
+	GString *send, *receive;
 	GnomeKeyringResult res;
 	
 	send = g_string_new (NULL);
@@ -910,10 +1014,21 @@ gnome_keyring_lock_sync (const char *keyring)
 		return GNOME_KEYRING_RESULT_BAD_ARGUMENTS;
 	}
 	
-	res = run_sync_operation (send, NULL);
+ 	receive = g_string_new (NULL);
+	res = run_sync_operation (send, receive);
 	g_string_free (send, TRUE);
-	
-	return res;
+	if (res != GNOME_KEYRING_RESULT_OK) {
+		g_string_free (receive, TRUE);
+		return res;
+	}
+
+        if (!gnome_keyring_proto_decode_result_reply (receive, &res)) {
+                g_string_free (receive, TRUE);
+                return GNOME_KEYRING_RESULT_IO_ERROR;
+        }
+        g_string_free (receive, TRUE);
+
+        return res;
 }
 
 gpointer
@@ -943,7 +1058,7 @@ gnome_keyring_delete (const char                                  *keyring,
 GnomeKeyringResult
 gnome_keyring_delete_sync (const char *keyring)
 {
-	GString *send;
+	GString *send, *receive;
 	GnomeKeyringResult res;
 	
 	send = g_string_new (NULL);
@@ -955,9 +1070,79 @@ gnome_keyring_delete_sync (const char *keyring)
 		return GNOME_KEYRING_RESULT_BAD_ARGUMENTS;
 	}
 	
-	res = run_sync_operation (send, NULL);
+ 	receive = g_string_new (NULL);
+	res = run_sync_operation (send, receive);
 	g_string_free (send, TRUE);
+	if (res != GNOME_KEYRING_RESULT_OK) {
+		g_string_free (receive, TRUE);
+		return res;
+	}
+
+        if (!gnome_keyring_proto_decode_result_reply (receive, &res)) {
+                g_string_free (receive, TRUE);
+                return GNOME_KEYRING_RESULT_IO_ERROR;
+        }
+        g_string_free (receive, TRUE);
+
+        return res;
+}
+
+gpointer
+gnome_keyring_change_password (const char                                  *keyring,
+		      const char                                  *original,
+		      const char                                  *password,
+		      GnomeKeyringOperationDoneCallback            callback,
+		      gpointer                                     data,
+		      GDestroyNotify                               destroy_data)
+{
+	GnomeKeyringOperation *op;
 	
+	op = start_operation (callback, CALLBACK_DONE, data, destroy_data);
+	if (op->state == STATE_FAILED) {
+		return op;
+	}
+	
+	if (!gnome_keyring_proto_encode_op_string_string_string (op->send_buffer,
+							  GNOME_KEYRING_OP_CHANGE_KEYRING_PASSWORD,
+							  keyring, original, password)) {
+		schedule_op_failed (op, GNOME_KEYRING_RESULT_BAD_ARGUMENTS);
+	}
+
+	op->reply_handler = gnome_keyring_standard_reply;
+	
+	return op;
+}
+
+GnomeKeyringResult
+gnome_keyring_change_password_sync (const char *keyring_name,
+			   const char *original, const char *password)
+{
+	GString *send, *receive;
+	GnomeKeyringResult res;
+	
+	send = g_string_new (NULL);
+	
+	if (!gnome_keyring_proto_encode_op_string_string_string (send,
+							  GNOME_KEYRING_OP_CHANGE_KEYRING_PASSWORD,
+							  keyring_name, original, password)) {
+		g_string_free (send, TRUE);
+		return GNOME_KEYRING_RESULT_BAD_ARGUMENTS;
+	}
+	
+ 	receive = g_string_new (NULL);
+	res = run_sync_operation (send, receive);
+	g_string_free (send, TRUE);
+	if (res != GNOME_KEYRING_RESULT_OK) {
+		g_string_free (receive, TRUE);
+		return res;
+	}
+
+	if (!gnome_keyring_proto_decode_result_reply (receive, &res)) {
+		g_string_free (receive, TRUE);
+		return GNOME_KEYRING_RESULT_IO_ERROR;
+	}
+	g_string_free (receive, TRUE);
+
 	return res;
 }
 
@@ -1014,8 +1199,9 @@ gnome_keyring_get_info_sync (const char        *keyring,
 	
 	*info = NULL;
 
-	if (!gnome_keyring_proto_encode_op_only (send,
-						 GNOME_KEYRING_OP_GET_KEYRING_INFO)) {
+	if (!gnome_keyring_proto_encode_op_string (send,
+						   GNOME_KEYRING_OP_GET_KEYRING_INFO,
+						   keyring)) {
 		g_string_free (send, TRUE);
 		return GNOME_KEYRING_RESULT_BAD_ARGUMENTS;
 	}
@@ -1066,7 +1252,7 @@ GnomeKeyringResult
 gnome_keyring_set_info_sync (const char       *keyring,
 			     GnomeKeyringInfo *info)
 {
-	GString *send;
+	GString *send, *receive;
 	GnomeKeyringResult res;
 	
 	send = g_string_new (NULL);
@@ -1077,8 +1263,10 @@ gnome_keyring_set_info_sync (const char       *keyring,
 		return GNOME_KEYRING_RESULT_BAD_ARGUMENTS;
 	}
 	
-	res = run_sync_operation (send, NULL);
+ 	receive = g_string_new (NULL);
+	res = run_sync_operation (send, receive);
 	g_string_free (send, TRUE);
+	g_string_free (receive, TRUE);
 	
 	return res;
 }
@@ -1160,6 +1348,38 @@ gnome_keyring_list_item_ids_sync (const char  *keyring,
 	return res;
 }
 
+GnomeKeyringResult
+gnome_keyring_daemon_set_display_sync (const char *display)
+{
+	GString *send, *receive;
+	GnomeKeyringResult res;
+
+	send = g_string_new (NULL);
+
+	if (!gnome_keyring_proto_encode_op_string (send,
+						   GNOME_KEYRING_OP_SET_DAEMON_DISPLAY,
+						   display)) {
+		g_string_free (send, TRUE);
+		return GNOME_KEYRING_RESULT_BAD_ARGUMENTS;
+	}
+
+	receive = g_string_new (NULL);
+	res = run_sync_operation (send, receive);
+	g_string_free (send, TRUE);
+	if (res != GNOME_KEYRING_RESULT_OK) {
+		g_string_free (receive, TRUE);
+		return res;
+	}
+
+	if (!gnome_keyring_proto_decode_result_reply (receive, &res)) {
+		g_string_free (receive, TRUE);
+		return GNOME_KEYRING_RESULT_IO_ERROR;
+	}
+	g_string_free (receive, TRUE);
+
+	return res;
+}
+
 void
 gnome_keyring_info_set_lock_on_idle (GnomeKeyringInfo *keyring_info,
 				     gboolean          value)
@@ -1314,6 +1534,17 @@ gnome_keyring_find_itemsv (GnomeKeyringItemType                  type,
 	return op;
 }
 
+/**
+ * gnome_keyring_find_items_sync:
+ * @type: a #GnomeKeyringItemType
+ * @attributes: a #GnomeKeyringAttributeList
+ * @found: a return location for the found items, must not be %NULL
+ *
+ * Find elements of type #GnomeKeyring by matching attributes and @type.
+ *
+ * Returns: a #GList of #GnomeKeyringFound, free with g_list_free() and
+ * gnome_keyring_found_free()
+ */
 GnomeKeyringResult
 gnome_keyring_find_items_sync (GnomeKeyringItemType        type,
 			       GnomeKeyringAttributeList  *attributes,
@@ -1481,7 +1712,7 @@ GnomeKeyringResult
 gnome_keyring_item_delete_sync (const char *keyring,
 				guint32     id)
 {
-	GString *send;
+	GString *send, *receive;
 	GnomeKeyringResult res;
 
 	send = g_string_new (NULL);
@@ -1494,8 +1725,10 @@ gnome_keyring_item_delete_sync (const char *keyring,
 		return GNOME_KEYRING_RESULT_BAD_ARGUMENTS;
 	}
 
-	res = run_sync_operation (send, NULL);
+ 	receive = g_string_new (NULL);
+	res = run_sync_operation (send, receive);
 	g_string_free (send, TRUE);
+	g_string_free (receive, TRUE);
 
 	return res;
 }
@@ -1610,7 +1843,7 @@ gnome_keyring_item_set_info_sync (const char           *keyring,
 				  guint32               id,
 				  GnomeKeyringItemInfo *info)
 {
-	GString *send;
+	GString *send, *receive;
 	GnomeKeyringResult res;
 	
 	send = g_string_new (NULL);
@@ -1621,8 +1854,10 @@ gnome_keyring_item_set_info_sync (const char           *keyring,
 		return GNOME_KEYRING_RESULT_BAD_ARGUMENTS;
 	}
 	
-	res = run_sync_operation (send, NULL);
+ 	receive = g_string_new (NULL);
+	res = run_sync_operation (send, receive);
 	g_string_free (send, TRUE);
+	g_string_free (receive, TRUE);
 	
 	return res;
 }
@@ -1756,7 +1991,7 @@ gnome_keyring_item_set_attributes_sync (const char                *keyring,
 					guint32                    id,
 					GnomeKeyringAttributeList *attributes)
 {
-	GString *send;
+	GString *send, *receive;
 	GnomeKeyringResult res;
 	
 	send = g_string_new (NULL);
@@ -1768,8 +2003,10 @@ gnome_keyring_item_set_attributes_sync (const char                *keyring,
 		return GNOME_KEYRING_RESULT_BAD_ARGUMENTS;
 	}
 	
-	res = run_sync_operation (send, NULL);
+ 	receive = g_string_new (NULL);
+	res = run_sync_operation (send, receive);
 	g_string_free (send, TRUE);
+	g_string_free (receive, TRUE);
 	
 	return res;
 
@@ -1869,7 +2106,7 @@ gnome_keyring_item_set_acl_sync (const char *keyring,
 				 guint32     id,
 				 GList      *acl)
 {
-	GString *send;
+	GString *send, *receive;
 	GnomeKeyringResult res;
 	
 	send = g_string_new (NULL);
@@ -1881,12 +2118,51 @@ gnome_keyring_item_set_acl_sync (const char *keyring,
 		return GNOME_KEYRING_RESULT_BAD_ARGUMENTS;
 	}
 	
-	res = run_sync_operation (send, NULL);
+ 	receive = g_string_new (NULL);
+	res = run_sync_operation (send, receive);
 	g_string_free (send, TRUE);
+	g_string_free (receive, TRUE);
 	
 	return res;
 }
 
+GnomeKeyringResult 
+gnome_keyring_item_grant_access_rights_sync (const char                   *keyring, 
+					     const char                   *display_name, 
+					     const char                   *full_path, 
+					     const guint32                id, 
+					     const GnomeKeyringAccessType rights) 
+{
+ 	GList *acl_list = NULL;
+ 	GnomeKeyringApplicationRef new_app_ref;
+ 	GnomeKeyringAccessControl acl;
+	GnomeKeyringResult res;
+
+	/* setup application structure */
+	new_app_ref.display_name = (char *) display_name;
+	new_app_ref.pathname = (char *) full_path;
+	acl.application = &new_app_ref; 
+	acl.types_allowed = rights; 
+
+	/* get the original acl list */
+	res = gnome_keyring_item_get_acl_sync (keyring,
+					       id,
+					       &acl_list);
+	if (GNOME_KEYRING_RESULT_OK != res)
+		goto out;
+
+	/* append access rights */
+	acl_list = g_list_append (acl_list, (gpointer) &acl);
+	res = gnome_keyring_item_set_acl_sync (keyring, 
+					       id,
+					       acl_list);
+out:
+	if (acl_list)
+		g_list_free (acl_list);
+
+	return res;
+}
+
 GnomeKeyringItemType
 gnome_keyring_item_info_get_type (GnomeKeyringItemInfo *item_info)
 {
@@ -2207,6 +2483,7 @@ gnome_keyring_find_network_password_sync (const char
 	GnomeKeyringResult result;
 	GList *found;
 	
+	*out_list = NULL;
 	attributes = make_attribute_list_for_network_password (user,
 							       domain,
 							       server,
